/* 
 * Copyright (C) 2006, iRex Technologies B.V.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Portions of this code are taken from xpdf and are
 * Copyright 1996-2003 Glyph & Cog, LLC
 */

#include "PDFCore.h"
#include "PDFViewerLog.h"
#include "controller.h"
#include <math.h>

CPDFCore::CPDFCore(SplashColorMode colorModeA, int bitmapRowPadA,
	         GBool reverseVideoA, SplashColorPtr paperColorA)
{
    outputDev = NULL;//new SplashOutputDev(colorModeA, bitmapRowPadA, reverseVideoA, paperColorA);
    textOut = NULL;//new TextOutputDev(NULL, gTrue, gFalse, gFalse);
    pdfdoc = NULL;    
    ctrl   = NULL;
}

CPDFCore::~CPDFCore(void)
{
    if (outputDev) { delete outputDev; outputDev = NULL; }
    if (pdfdoc) { delete pdfdoc; pdfdoc = NULL; }
    ctrl = NULL;
}

// we can reuse this core for multiple documents
int CPDFCore::open(GooString *fileNameA, GooString *ownerPassword,
         GooString *userPassword , void *guiDataA )
{
    int ret = errNone;
    // already open 
    if (pdfdoc)
    {
        PV_LOGPRINTF("Delete old doc %p\n", pdfdoc);
        delete pdfdoc;
    }

    // construct a new one
    pdfdoc = new PDFDoc(fileNameA, ownerPassword, userPassword, guiDataA);
    
    // check result
    if (!pdfdoc->isOk())
    {
        ret = pdfdoc->getErrorCode();
        delete pdfdoc;
        PV_ERRORPRINTF("Open PDF document %s failed with error code %d", 
            fileNameA->getCString(), ret);                
        return errOpenFile; 
    }
    
    // done
    if (NULL == outputDev)
    {
        SplashColor     white;
        white[0] = 255;
        white[1] = 255;
        white[2] = 255;
        outputDev = new SplashOutputDev(splashModeMono8, 1, gFalse, white);
    }
    outputDev->startDoc(pdfdoc->getXRef());
    return ret;
}

int CPDFCore::open(BaseStream *stream, GooString *ownerPassword ,
         GooString *userPassword , void *guiDataA)
{
    return errNone;
}

// only close the document
void CPDFCore::close()
{
    if (outputDev) {  delete outputDev; outputDev = NULL; }
    if (pdfdoc) { delete pdfdoc; pdfdoc = NULL; }
}

void CPDFCore::getPageBasicInfo(CPageInfo & info)
{
    info.pageDx = pdfdoc->getPageCropWidth(info.pageNumber);
    info.pageDy = pdfdoc->getPageCropHeight(info.pageNumber);
    info.pageRotate = pdfdoc->getPageRotate(info.pageNumber);
}

int CPDFCore::getNumPages()
{
    if (NULL == pdfdoc) return -1;
    return pdfdoc->getNumPages();
}

RenderRet CPDFCore::renderPage(CPageInfo & info, GBool useMediaBox, 
                     GBool crop, GBool doLinks, GBool (*abortCheckCbk)(void *data) ,
                     void *abortCheckCbkData )
{
    if (NULL == pdfdoc)
    {
        PV_ERRORPRINTF("PDF document has not been opened!");    
        return Render_Error;
    }

    if (0 >= info.pageNumber || info.pageNumber > pdfdoc->getNumPages())
    {
        PV_ERRORPRINTF("PageNumber %d is invalid![1, %d]", 
            info.pageNumber,  pdfdoc->getNumPages());
        return Render_Error;
    }

    // get page basic information
    getPageBasicInfo(info);

    // calcuate layout
    calculateLayout(info);

    // render it now
#if (PV_PROFILE_ON)
    int t1 = ctrl->getHighTimeStamp();
#endif 
    RenderRet ret = pdfdoc->displayPage(outputDev, info.pageNumber, 
        info.hDPI, info.vDPI, info.rotate, useMediaBox, crop, doLinks,
        abortCheckCbk, abortCheckCbkData);
    
    // check result
    if (ret == Render_Error || ret == Render_Invalid)
    {
        PV_ERRORPRINTF("Render page error!");
        return ret;
    }

    // bitmap has been allocated
    info.bitmap = outputDev->takeBitmap();
    if (ret == Render_Done)
    {
#if (PV_PROFILE_ON)
        int t2 = ctrl->getHighTimeStamp();
        PV_DUMP("PDFDoc::displayPage uses %d\n", t2 - t1);
#endif 

#if (PV_PROFILE_ON)
        t1 = ctrl->getHighTimeStamp();
        PV_DUMP("xMgr::createImage uses %d\n", t1 - t2);
#endif 
        
        // copy ctm and ictm
        memcpy(info.ctm, outputDev->getDefCTM(), 6 * sizeof(double));
        memcpy(info.ictm, outputDev->getDefICTM(), 6 * sizeof(double));        
                
        // take links
        info.links  = pdfdoc->takeLinks();
        
        // take text
        if (textOut)
        {
            pdfdoc->displayPage(textOut, info.pageNumber, info.hDPI, info.vDPI, info.rotate,
                gFalse, gTrue, gFalse);
            info.text = textOut->takeText();                
        }
        return Render_Done;
    }
    else if (ret == Render_Abort)
    {
        delete info.bitmap; 
        info.bitmap = NULL;
        info.links  = NULL;
    }
    return Render_Abort;
}

// calculate the layout according to zoom and rotation
void CPDFCore::calculateLayout(CPageInfo & info)
{
    int rot = info.rotate + info.pageRotate;
    if (rot >= 360) rot -= 360;
    else if (rot < 0) rot += 360;
    double uw = info.pageDx, uh = info.pageDy, ut;
    if (rot == 90 || rot == 270)
    {
        ut = uw; uw = uh; uh = ut;
    }

    // when ZoomPage, should select DPI smaller 
    if (info.pageZoom == ZoomPage)
    {
        info.hDPI = ( ctrl->settings.drawAreaWidth / uw) * DefDPI;
        info.vDPI = ( ctrl->settings.drawAreaHeight / uh) * DefDPI;
        if (info.hDPI < info.vDPI) 
        {
            info.vDPI = info.hDPI;
        }
        else
        {
            info.hDPI = info.vDPI;
        }
    }
    else if (info.pageZoom == ZoomWidth)
    {
        if (ctrl->settings.rotate == 0 || ctrl->settings.rotate == 180)
        {   
            info.hDPI = ( ctrl->settings.drawAreaWidth / uw) * DefDPI;
            info.vDPI = info.hDPI;
        }
        else if (ctrl->settings.rotate == 90 || ctrl->settings.rotate == 270)            
        {
            info.hDPI = ( ctrl->settings.drawAreaHeight / uh) * DefDPI;
            info.vDPI = info.hDPI;
        }
    }
    else
    {
        info.hDPI = info.pageZoom * 0.01 * DefDPI;
        info.vDPI = info.pageZoom * 0.01 * DefDPI;
    }
}

// http://www.paperonweb.com/size.htm
int CPDFCore::getPageType(const int pageNumber)
{
    if (NULL == pdfdoc) 
    {
        PV_ERRORPRINTF("Try to access NULL pdfdoc pointer!");
        return -1;
    }
    if (pageNumber < 1 || pageNumber > pdfdoc->getNumPages()) 
    {
        PV_ERRORPRINTF("Invalid page number %d!", pageNumber);
        return -1;
    }
    double pageMx = pdfdoc->getPageMediaWidth(pageNumber);
    double pageMy = pdfdoc->getPageMediaHeight(pageNumber);
    double hISO, wISO;
    if ((fabs(pageMx - 612) < 0.1 && fabs(pageMy - 792) < 0.1) ||
    	(fabs(pageMx - 792) < 0.1 && fabs(pageMy - 612) < 0.1)) 
    {
        return 41; // letter but not A4         
    } 
    else 
    {
        hISO = sqrt(sqrt(2.0)) * 7200 / 2.54;
        wISO = hISO / sqrt(2.0);
        for (int i = 0; i <= 9; ++i) 
        {
	        if ((fabs(pageMx - wISO) < 1 && fabs(pageMy - hISO) < 1) ||
	            (fabs(pageMx - hISO) < 1 && fabs(pageMy - wISO) < 1)) 
            {
                return i * 10;
            }
            
            // not correct, used to compare 
            if ((pageMx > wISO && pageMy > hISO) ||
	            (pageMx > hISO && pageMy > wISO))
            {
                return i * 10 + 5; 
            }	            
    	    hISO = wISO;
    	    wISO /= sqrt(2.0);
        }    	    
    }
    return -1;      
}

double CPDFCore::getZoomMax(const int pageNumber)
{
    int type = getPageType(pageNumber);
    switch (type)
    {
    case 0:
    case 5:
        PV_LOGPRINTF("Page A0");
        return A0ZoomMax;
    case 10:
    case 15:
        PV_LOGPRINTF("Page A1");    
        return A1ZoomMax;        
    case 20:
    case 25:
        PV_LOGPRINTF("Page A2");    
        return A2ZoomMax;
    case 30:
    case 35:
        PV_LOGPRINTF("Page A3");    
        return A3ZoomMax;
    case 40:
    case 45:
        PV_LOGPRINTF("Page A4");    
        return A4ZoomMax;
    case 50:
    case 55:
        PV_LOGPRINTF("Page A5");    
        return A5ZoomMax;
    case 60:
    case 65:
        PV_LOGPRINTF("Page A6");    
        return A6ZoomMax;         
    case 70:
    case 75:
        PV_LOGPRINTF("Page A7");                                       
        return A7ZoomMax;
    case 80:
    case 85:
        PV_LOGPRINTF("Page A8");                                       
        return A8ZoomMax;
    case 90:
    case 95:
        PV_LOGPRINTF("Page A9");                                       
        return A9ZoomMax;
    default:
        PV_LOGPRINTF("Page Unknown");
        return ZoomMax;        
    }
    return ZoomMax;
}

//////////////////////////////////////////////////////////////////////////
// verifyZoomMax compares zoom value with max value the page can be displayed
//////////////////////////////////////////////////////////////////////////
double CPDFCore::verifyZoomMax(double zoom,const int pageNumber)
{
    double tmp = getZoomMax(pageNumber);
    if (zoom > tmp)
    {
        return tmp;
    }
    return zoom;
}
