/* 
 * Copyright (C) 2006, iRex Technologies B.V.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Portions of this code are taken from xpdf and are
 * Copyright 1996-2003 Glyph & Cog, LLC
 */

#include "RenderThread.h"
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include "controller.h"
#include "PDFViewerLog.h"
#include <unistd.h>

#define Object XtObject
#include <Xm/XmAll.h>
#undef Object
#ifdef LESSTIF_VERSION
#undef XtDisplay
#undef XtScreen
#undef XtWindow
#undef XtParent
#undef XtIsRealized
#endif

CRenderThread::CRenderThread(void)
: bRunning(gFalse)
{
    initialize();
}

CRenderThread::~CRenderThread(void)
{
    destroy();
}

CRenderThread::threadState CRenderThread::abortState = Initialized;

void CRenderThread::initialize()
{
    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);
    pthread_mutex_init(&abort, NULL);
    renderingPage = NULL;
}

void CRenderThread::destroy()
{
    pthread_mutex_destroy(&mutex);
    pthread_cond_destroy(&cond);
    pthread_mutex_destroy(&abort);
}

GBool CRenderThread::abortCheckFunc(void *data)
{
    if (abortState == ToAbort)
    {
        // PV_LOGPRINTF("Abort page!");
        return gTrue;
    }
    return gFalse;
    
}

void CRenderThread::start(ThreadParam * p)
{
    if (bRunning) 
    {
        PV_ERRORPRINTF("Thread already running!");
        return;
    }
    param = p;
    param->bQuit = gFalse;
    param->thread = this;
    renderingPage = NULL;
    abortState = Initialized;
    if (pthread_create(&thread, NULL, threadProc, p) != 0)
    {
        PV_ERRORPRINTF("Could not create thread!");
        return;
    }
    PV_LOGPRINTF("Create render thread done!");
    bRunning = gTrue;
}

void CRenderThread::stop()
{
    if (!bRunning) return;

    abortState = Initialized;
    param->bQuit = gTrue;
    pthread_mutex_lock(&mutex);
    pthread_cond_signal(&cond);
    pthread_mutex_unlock(&mutex);
    pthread_join(thread, NULL);
    bRunning = gFalse;
    renderingPage = NULL;
}

void CRenderThread::addTodoItem(TodoItem * item, GBool bHead)
{
    param->todoList->pushItem(item, bHead);
}

void CRenderThread::signal()
{
    pthread_mutex_lock(&mutex);
    pthread_cond_signal(&cond);
    pthread_mutex_unlock(&mutex);
}

void CRenderThread::adjustPageList()
{
    // memory       allocate         pages
    // 50MB  ------- 20 MB   ------   35 
    // 30MB  ------- 30 MB   ------    8 ( 250%)
    static const int MEM = 1024 * 1024 * 20;    // 49.1 MB used from experiment, ~ 35 pages cached
    while (1)
    {
        char * p = (char *)malloc(MEM);
        if (p)
        {
            PV_MEM("Memory is enough %d bytes!", MEM);
            free(p);
            return;
        }
        else 
        {
            PV_MEM("Not enough memory, try to release some pages!");
            for(int i = 0; i < RemovePages; ++i)
            {
                param->pageList->removeOldestOne();
            }                
        }
    }
}

void CRenderThread::renderPageNow(TodoItem * item)
{
    // if it's already rendering ignore the request
    pthread_mutex_lock(&abort);
    if (abortState == Rendering && 
        renderingPage &&
        renderingPage->isSame(item->pageNumber, item->zoom, item->rotate))
    {
        PV_LOGPRINTF("Page %d is in rendering!", item->pageNumber);
        delete item;
        pthread_mutex_unlock(&abort);
        return;
    }        

    if (abortState == Rendering)
    {
        abortState = ToAbort;
    }
    param->todoList->pushItem(item, gTrue);
    signal();
    pthread_mutex_unlock(&abort);
}

// ugly, I have to reuse KeyPress. 
// PDFView has to check the time field
static void postEvent(ThreadParam * param, CPageInfo *page)
{
    XKeyEvent xev;
    xev.type = KeyPress;
    xev.display = param->ctrl->xMgr.display;
    xev.root = RootWindow(param->ctrl->xMgr.display, param->ctrl->xMgr.screenNum);
    xev.window = XtWindow(param->ctrl->xMgr.win);
    xev.subwindow = None;
    xev.time = 0;
    xev.x = page->pageNumber;
    xev.y = 1;
    xev.x_root = -1;
    xev.y_root = -1;
    xev.same_screen = 1;
    xev.state = 0;
    xev.keycode = XKeysymToKeycode(param->ctrl->xMgr.display, 0x0000);
    XSendEvent(param->ctrl->xMgr.display, 
        XtWindow(param->ctrl->xMgr.win), 1, KeyPressMask, (XEvent *)&xev);
    XFlush(xev.display);
}

//////////////////////////////////////////////////////////////////////////
// rendering thread
// 1. check quit flag
// 2. check todo list
// 3. check page list
// 4. check memory
// 5. construct page item
// 6. record current rendering page
// 7. 
void * CRenderThread::threadProc(void * pvoid)
{
    // param
    ThreadParam * param = (ThreadParam *)pvoid;

    // loop
    for(;;)
    {
        if (param->bQuit)
        {
            PV_LOGPRINTF("Quit from thread!");
            return NULL;
        }
        
        // check todo list
        TodoItem *pItem = param->todoList->popItem();
        if (NULL == pItem) 
        {
            pthread_mutex_lock(&param->thread->mutex);
            PV_LOGPRINTF("Waiting...");
            pthread_cond_wait(&param->thread->cond, &param->thread->mutex); 
            pthread_mutex_unlock(&param->thread->mutex);
            continue;   // quit or do something 
        }
    
        // check page list 
        CPageInfo * page = param->pageList->checkPage(pItem);
        if (page) 
        {
            delete pItem;
            continue;
        }
        
        // memory control
        param->thread->adjustPageList();
                    
        // construct a new page item
        page = new CPageInfo;
        page->pageNumber = pItem->pageNumber;
        page->pageZoom   = pItem->zoom;
        page->rotate     = pItem->rotate;   // global settings
        delete pItem;
        
        // protect, renderingPage could be accessed by main thread and helper thread in same time
        pthread_mutex_lock(&param->thread->abort);
        param->thread->renderingPage = page;
        param->thread->abortState = Rendering;
        pthread_mutex_unlock(&param->thread->abort);
        
        // go
        RenderRet ret = param->ctrl->core->renderPage(*page, 
                gFalse, gTrue, gTrue, abortCheckFunc, NULL);
    
        // reset                
        pthread_mutex_lock(&param->thread->abort);
        param->thread->renderingPage = NULL;
        param->thread->abortState = RenderFinished;
        pthread_mutex_unlock(&param->thread->abort);                                        

        // check result
        if (ret != Render_Done)
        {
            if (ret == Render_Abort)
            {
                PV_LOGPRINTF("Page %d is aborted!", page->pageNumber);
            }
            delete page;
            continue;
        }
        page->timestamp = param->ctrl->getTimeStamp();
        param->pageList->addPage(page);

        // send a message to the main window
        postEvent(param, page);
    }
    return NULL;
}
